/*
 *                      
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 * A SLAM driver based on gmapping
 */

#include <unistd.h>
#include <string.h>
#include <ctime>
#include <cmath>
#include <libplayercore/playercore.h>

#include <libplayercore/configfile.h>

#include "slamdriver.h"
#include "point.h"

inline int _3d_to_1d(int cur_row, int cur_col, int cur_color, int max_row, int max_col, int max_color)
{
//    return max_color*cur_row+max_color*cur_col*max_row+cur_color;
    return max_color*cur_col + max_color*cur_row*max_col + cur_color;
}

inline int _2d_to_1d(int cur_row, int cur_col, int max_row, int max_col)
{
    return cur_row*max_col+cur_col;
}
    
using namespace GMapping;

// Driver creation function. In this function, we create and return
// (as a generic Driver*) a pointer to a new instance of this driver.
Driver* SlamDriver_Init(ConfigFile* cf, int section)
{
    // Create and return a new instance of this driver
    return((Driver*)(new SlamDriver(cf, section)));
}

// Driver registration function In this function, we add
// the driver into the given driver table, indicating which interface the
// driver can support and how to create a driver instance.
void SlamDriver_Register(DriverTable* table)
{
    table->AddDriver("slam", SlamDriver_Init);
}

////////////////////////////////////////////////////////////////////////////////
// Constructor.
SlamDriver::SlamDriver(ConfigFile* cf, int section) : Driver(cf, section) 
{

    PLAYER_MSG0(4, "Starting SLAM constructor");

    memset(&this->localize_addr, 0, sizeof(player_devaddr_t));
    memset(&this->position_addr, 0, sizeof(player_devaddr_t));
    memset(&this->map_addr     , 0, sizeof(player_devaddr_t));
    memset(&this->camera_addr  , 0, sizeof(player_devaddr_t));
    
    // Do we create a localize interface?
    if(cf->ReadDeviceAddr(&(this->localize_addr), section, "provides", PLAYER_LOCALIZE_CODE, -1, NULL) == 0)
    {
        if(this->AddInterface(this->localize_addr))
        {
            this->SetError(-1);
            return;
        }
        PLAYER_MSG0(4, "Provides localize interface");
    }

    // Do we create a position interface?
    if(cf->ReadDeviceAddr(&(this->position_addr), section, "provides", PLAYER_POSITION2D_CODE, -1, NULL) == 0)
    {
        if(this->AddInterface(this->position_addr))
        {
            this->SetError(-1);
            return;
        }
        PLAYER_MSG0(4, "Provides position2d interface");
    }

    // Do we create a map interface?
    if(cf->ReadDeviceAddr(&(this->map_addr), section, "provides", PLAYER_MAP_CODE, -1, NULL) == 0)
    {
        if(this->AddInterface(this->map_addr))
        {
            this->SetError(-1);
            return;
        }
        PLAYER_MSG0(4, "Provides map interface");
    }

    // Do we create a camera interface?
    if(cf->ReadDeviceAddr(&(this->camera_addr), section, "provides", PLAYER_CAMERA_CODE, -1, NULL) == 0)
    {
        if(this->AddInterface(this->camera_addr))
        {
            this->SetError(-1);
            return;
        }
        PLAYER_MSG0(4, "Provides camera interface");
    }

    memset(&this->odom_addr, 0, sizeof(player_devaddr_t));
    memset(&this->laser_addr,0, sizeof(player_devaddr_t));

    // Create odometry sensor
    this->odom = NULL;
    if(cf->ReadDeviceAddr(&this->odom_addr, section, "requires", PLAYER_POSITION2D_CODE, -1, NULL) != 0)
    {
        PLAYER_ERROR("SLAM needs an odometry sensor");
        this->SetError(-1);
        return;
    }
    PLAYER_MSG4(2, "Odometry address:: Host:%d, robot:%d, interface:%d, index:%d", odom_addr.host, odom_addr.robot, odom_addr.interf, odom_addr.index);

    // Create laser sensor
    this->laser = NULL;
    if (cf->ReadDeviceAddr(&this->laser_addr, section, "requires", PLAYER_LASER_CODE, -1, NULL) != 0)
    {
        PLAYER_ERROR("SLAM needs a laser sensor");
        this->SetError(-1);
        return;
    }
    PLAYER_MSG4(2, "Laser address:: Host:%d, robot:%d, interface:%d, index:%d", laser_addr.host, laser_addr.robot, laser_addr.interf, laser_addr.index);
    
    if (!this->laser_addr.interf) 
    {
        PLAYER_ERROR("SLAM needs exactly a laser sensor");
        this->SetError(-1);
        return;
    }

    //////////////////////////////////////
    // Read Configuration options       //
    //////////////////////////////////////
    
    this->particles                         = cf->ReadInt(section, "particles", 10); // number of particles
    // gfs measurement integration
    this->angular_update                    = cf->ReadAngle(section, "angular_update", 1.0); // angular threshold to update map
    this->linear_update                     = cf->ReadLength(section, "linear_update", 0.5); // linear threshold to update map
    // map resolution
    this->map_delta                         = cf->ReadLength(section, "map_delta", 0.025);  // map resolution
    // TODO: Check if we can get this values from the laser device
    // scan matcher 
//    this->scan_matcher_max_range            = cf->ReadLength(section, "scan_matcher_max_range", 4); // (maximum valid) for URG Laser=4m
//    this->scan_matcher_max_urange           = cf->ReadLength(section, "scan_matcher_max_urange", 4); // (use up to)
    this->scan_matcher_sigma                = cf->ReadLength(section, "scan_matcher_sigma", 0.05); // scan matcher cell sigma for the greedy search
    this->scan_matcher_reg_score            = cf->ReadInt(section, "scan_matcher_reg_score", 10000); // minimum score for registering a scan
    this->scan_matcher_iterations           = cf->ReadInt(section, "scan_matcher_iterations", 5); // iterations
    this->scan_matcher_crit_score           = cf->ReadFloat(section, "scan_matcher_crit_score", 0.0); // critical score (leave this)
    this->scan_matcher_max_move             = cf->ReadLength(section, "scan_matcher_max_move", 1.0); // maximum move among two scans. This detects some corrupted logs
    this->scan_matcher_autosize             = cf->ReadBool(section, "scan_matcher_autosize", "no"); // determine the map size by pre readoing the log
    
    this->scan_matcher_linear_step          = cf->ReadLength(section, "scan_matcher_linear_step", 0.05); // linear search step (choose delta)
    this->scan_matcher_angular_step         = cf->ReadAngle(section, "scan_matcher_angular_step", 0.05); // angular search step, this is fine, depending on the odometry error and the update interval
    this->scan_matcher_like_sigma           = cf->ReadFloat(section, "scan_matcher_like_sigma", 0.075); // sigma likelihood of 1 beam
    this->scan_matcher_like_skip            = cf->ReadInt(section, "scan_matcher_like_skip", 1); // beams to skip in the likelihood computation
    this->scan_matcher_skip_matching        = cf->ReadBool(section, "scan_matcher_skip_matching", "no"); // do not perform scan matching before computing the statistics
    
    this->scan_matcher_kernel_size          = cf->ReadInt(section, "scan_matcher_kernel_size", 1.0); // the higher the value the slower the filter, the better it can deal with noise, but the less precise and slower
    this->scan_matcher_gain_smooth          = cf->ReadInt(section, "scan_matcher_gain_smooth", 3); // gain for smoothing the likelihood
    this->scan_matcher_resample_threshold   = cf->ReadFloat(section, "scan_matcher_resample_threshold", 0.5); // when neff is below this value a resampling occurs
    // likelihood sampling
    this->like_sampling_linear_range        = cf->ReadLength(section, "like_sampling_linear_range", 0.2); // linear range
    this->like_sampling_linear_step         = cf->ReadLength(section, "like_sampling_linear_step", 0.2); // linear step
    this->like_sampling_angular_range       = cf->ReadAngle(section, "like_sampling_angular_range", 0.05); // angular range
    this->like_sampling_angular_step        = cf->ReadAngle(section, "like_sampling_angular_step", 0.05); // angular step
    // motion model parameters
    this->motion_model_trans_fun_trans      = cf->ReadFloat(section, "motion_model_translation_function_translation", 0.1); // translation as a function of translation
    this->motion_model_trans_fun_rot        = cf->ReadFloat(section, "motion_model_translation_function_rotation", 0.1); // translation as a function of rotation
    this->motion_model_rot_fun_trans        = cf->ReadFloat(section, "motion_model_rotation_function_translation", 0.1); // rotation as a function of translation
    this->motion_model_rot_fun_rot          = cf->ReadFloat(section, "motion_model_rotation_function_rotation", 0.1); // rotation as a function of rotation
    // odometry integration in proposal
    this->linear_odometry_reliability       = cf->ReadFloat(section, "linear_odometry_reliability", 0.0); // linear odometry reliability
    this->angular_odometry_reliability      = cf->ReadFloat(section, "angular_odometry_reliability", 0.0); // angular odometry reliability
    this->consider_odometry_covariance      = cf->ReadBool(section, "consider_odometry_covariance", "no"); // consider odometry covariance
    // inital map params
    this->initial_map_x_min                 = cf->ReadLength(section, "initial_map_x_min", -6.0); // minimum x for initial map
    this->initial_map_y_min                 = cf->ReadLength(section, "initial_map_y_min", -6.0); // minimum y for initial map
    this->initial_map_x_max                 = cf->ReadLength(section, "initial_map_x_max", 6.0); // maximum x for initial map
    this->initial_map_y_max                 = cf->ReadLength(section, "initial_map_y_max", 6.0); // maximum y for initial map
    // initial robot position
    //     X
    //     ^
    //     |
    //Y <--+
//    this->initial_robot_x                   = cf->ReadLength(section, "initial_robot_x", 0.0); // initial robot position on the X axis
//    this->initial_robot_y                   = cf->ReadLength(section, "initial_robot_y", 0.0); // initial robot position on the Y axis
//    this->initial_robot_theta               = cf->ReadAngle(section, "initial_robot_theta", 0.0); // initial robot orientation

    this->generate_map                      = cf->ReadBool(section, "generate_map", "yes"); // Generate a map?
    this->write_map_to_file                 = cf->ReadBool(section, "write_map_to_file", "yes"); // Write the map to a file?
    this->map_filename                      = cf->ReadFilename(section, "map_filename", "cur_map.ppm"); // The map filename

    //////////////////////////////////
    // Set the pointers to NULL     //
    //////////////////////////////////
    frontLaser=NULL;
    cur_map=NULL;
    camera_data=NULL;
        
    PLAYER_MSG0(4, "SLAM constructor finished");

    return;
}

////////////////////////////////////////////////////////////////////////////////
// Set up the device.  Return 0 if things go well, and -1 otherwise.
int SlamDriver::Setup()
{   
    PLAYER_MSG0(4,"Slam driver setting up");

    // Setup the required devices
    if(!(this->laser = deviceTable->GetDevice(this->laser_addr)))
    {
        PLAYER_ERROR("unable to locate suitable laser device");
        return -1;
    }
    if (this->laser->Subscribe(this->InQueue) != 0)
    {
        PLAYER_ERROR("unable to subscribe to laser device");
        return -1;
    }

    if(!(this->odom = deviceTable->GetDevice(this->odom_addr)))
    {
        PLAYER_ERROR("unable to locate suitable position device");
        return -1;
    }
    if(this->odom->Subscribe(this->InQueue) != 0) //this->InQueue) != 0)
    {
        PLAYER_ERROR("unable to subscribe to position device");
        return -1;
    }
    
    // Request the configuration parameters from the laser device
    Message* msg;
	if (!(msg=this->laser->Request (this->InQueue, PLAYER_MSGTYPE_REQ, PLAYER_LASER_REQ_GET_CONFIG, NULL, 0, NULL,false)) ||
	     (msg->GetHeader()->size != sizeof(player_laser_config_t)) )
    {
        PLAYER_ERROR("failed to get configuration of underlying laser device");
        if(msg)
          delete msg;
        return(-1);
    }
    player_laser_config_t* lcfg = (player_laser_config_t*)msg->GetPayload();
    laser_config.min_angle  =  lcfg->min_angle;
    laser_config.max_angle  =  lcfg->max_angle;
    laser_config.resolution =  lcfg->resolution;
    laser_config.max_range  =  lcfg->max_range;
    laser_config.range_res  =  lcfg->range_res;
    laser_config.intensity  =  lcfg->intensity;
    
    PLAYER_MSG6(1, "Laser Config Struct: min angle:%.3f, max angle:%.3f, resolution: %.3f, max range:%.3f, range resolution:%.3f, intensity: %i", 
    laser_config.min_angle, laser_config.max_angle, laser_config.resolution, laser_config.max_range, laser_config.range_res, laser_config.intensity);
    
    // Request the geometry for the laser device
//    Message* msg;
	if (!(msg=this->laser->Request (this->InQueue, PLAYER_MSGTYPE_REQ, PLAYER_LASER_REQ_GET_GEOM, NULL, 0, NULL,false)) ||
	     (msg->GetHeader()->size != sizeof(player_laser_geom_t)) )
    {
        PLAYER_ERROR("failed to get geometry of underlying laser device");
        if(msg)
          delete msg;
        return(-1);
    }
    player_laser_geom_t* lgeom = (player_laser_geom_t*)msg->GetPayload();
    laser_pose3d=lgeom->pose;
    laser_size3d=lgeom->size;
    PLAYER_MSG6(1, "Laser Geom Struct: 3D Pose=(X:%.3f,Y:%.3f,Z:%.3f,Roll:%.3f,Pitch:%.3f,Yaw:%.3f)", 
                    laser_pose3d.px, laser_pose3d.py, laser_pose3d.pz, laser_pose3d.proll, laser_pose3d.ppitch, laser_pose3d.pyaw);
    PLAYER_MSG3(1, "Laser Geom Struct: 3D Size=(width:%.3f,length:%.3f,height:%.3f)", 
                    laser_size3d.sw, laser_size3d.sl, laser_size3d.sh);

    //////////////////////////////////
    // Create the gridSlamProcessor //
    ////////////////////////////////// 
    PLAYER_MSG0(1,"Creating a gridSlamProcessor instance");

    gmappingLogStream.open("../data/gmapping.log");

    gridSlamProcessor = new GridSlamProcessor(gmappingLogStream);
    gridSlamProcessor->setMatchingParameters(laser_config.max_range, laser_config.max_range, scan_matcher_sigma, 
                                             scan_matcher_kernel_size, scan_matcher_linear_step, scan_matcher_angular_step, 
                                             scan_matcher_iterations, scan_matcher_like_sigma, scan_matcher_gain_smooth, scan_matcher_like_skip);

    gridSlamProcessor->setMotionModelParameters(motion_model_trans_fun_trans, motion_model_trans_fun_rot, motion_model_rot_fun_trans, motion_model_rot_fun_rot);
    gridSlamProcessor->setUpdateDistances(linear_update, angular_update, scan_matcher_resample_threshold);
    gridSlamProcessor->setgenerateMap(generate_map);

    //////////////////////////////////
    // Create the RangeSensor class //
    //////////////////////////////////
    // The remaining initialization needs to go to the process message as that's where we find out about the laser capabilities
    
    PLAYER_MSG0(4,"Slam driver set UP");

    // Start the device thread; spawns a new thread and executes
    // ExampleDriver::Main(), which contains the main loop for the driver.
    StartThread();

    return(0);
}


////////////////////////////////////////////////////////////////////////////////
// Shutdown the device
int SlamDriver::Shutdown()
{
    PLAYER_MSG0(4,"Shutting Slam driver down");

    // Stop and join the driver thread
    StopThread();

    this->odom->Unsubscribe(this->InQueue);
    this->laser->Unsubscribe(this->InQueue);    

    // TODO: Check if we need to clean dynamically allocated memory and if we have any memory leaks  
    delete gridSlamProcessor;
    if (!frontLaser)
        delete frontLaser;
    frontLaser=NULL;
    if (!cur_map)
        delete cur_map;

//    if (!camera_data)
    camera_data=NULL;
        
      
    PLAYER_MSG0(4,"Slam driver has been shutdown");

    return(0);
}


int SlamDriver::ProcessMessage(QueuePointer & resp_queue, 
                                  player_msghdr * hdr,
                                  void * data)
{
    //////////////////////////
    // LOCALIZE INTERFACE   //
    //////////////////////////
    // Is it a request to set the filter's pose?
    if(Message::MatchMessage(hdr, PLAYER_MSGTYPE_REQ, PLAYER_LOCALIZE_REQ_SET_POSE, this->localize_addr))
    {
//        player_localize_set_pose_t* setposereq;
        // Send an ACK
        this->Publish(this->localize_addr, resp_queue, PLAYER_MSGTYPE_RESP_ACK, PLAYER_LOCALIZE_REQ_SET_POSE);
        return(0);
    }
    // Is it a request for the current particle set?
    else if(Message::MatchMessage(hdr, PLAYER_MSGTYPE_REQ, PLAYER_LOCALIZE_REQ_GET_PARTICLES, this->localize_addr))
    {
        player_localize_get_particles_t resp;
        memset(&resp, 0, sizeof(player_localize_get_particles_t));
        this->Publish(this->localize_addr, resp_queue, PLAYER_MSGTYPE_RESP_ACK, PLAYER_LOCALIZE_REQ_GET_PARTICLES, (void*)&resp);
        return(0);
    }
    //////////////////////////////////
    // POSITION 2D OUTPUT INTERFACE //
    //////////////////////////////////
    // Is it a request to report the geometry of the position 2d interface?
/*    else if(Message::MatchMessage(hdr, PLAYER_MSGTYPE_REQ, PLAYER_POSITION2D_REQ_GET_GEOM, position_addr))
    {
        player_position2d_geom_t geom={{0}};
        // just return a point so we don't get errors from playerv
        geom.size.sl = 0.01;
        geom.size.sw = 0.01;

        Publish(this->position_addr, resp_queue, PLAYER_MSGTYPE_RESP_ACK, PLAYER_POSITION2D_REQ_GET_GEOM, &geom, sizeof(geom), NULL);
        return(0);
    }
*/    // Is it an unhandled position 2d request?
    else if(Message::MatchMessage(hdr, PLAYER_MSGTYPE_REQ, -1, this->position_addr))
    {
        // Pass the request on to the underlying position device and wait for
        // the reply.
        Message* msg;

        if(!(msg = this->odom->Request(this->InQueue, hdr->type, hdr->subtype, (void*)data, hdr->size, &hdr->timestamp)))
        {
            PLAYER_WARN1("failed to forward config request with subtype: %d\n", hdr->subtype);
            return(-1);
        }
        player_msghdr_t* rephdr = msg->GetHeader();
        void* repdata = msg->GetPayload();
        // Copy in our address and forward the response
        rephdr->addr = this->position_addr;
        this->Publish(resp_queue, rephdr, repdata);
        delete msg;
        return(0);
    }
    // Is it an unhandled position 2d command?
    // TODO: The behaviour is almost perfect. 
    else if(Message::MatchMessage(hdr, PLAYER_MSGTYPE_CMD, -1, this->position_addr))
    {
        // Pass the command on to the underlying position device
        player_msghdr_t newhdr = *hdr;
        newhdr.addr = this->odom_addr;
    
        this->odom->PutMsg(this->InQueue, &newhdr, (void*)data);
        return(0);
    }
    //////////////////////////////////
    // POSITION 2D INPUT INTERFACE  //
    //////////////////////////////////
    // Is it a data packet? 
    else if(Message::MatchMessage(hdr, PLAYER_MSGTYPE_DATA, PLAYER_POSITION2D_DATA_STATE, this->odom_addr))
    {
        assert(hdr->size == sizeof(player_position2d_data_t));
//        PLAYER_MSG0(4, "position2d message received");
        ProcessOdom(hdr, *reinterpret_cast<player_position2d_data_t *> (data));
        return 0;
    }
    //////////////////////////////
    // LASER INPUT INTERFACE    //
    //////////////////////////////
    else if(Message::MatchMessage(hdr, PLAYER_MSGTYPE_DATA, PLAYER_LASER_DATA_SCAN, this->laser_addr))
    {
        // It's not always that big...
        //assert(hdr->size == sizeof(player_laser_data_t));
//        PLAYER_MSG0(4, "laser message received");
        ProcessLaser(hdr, *reinterpret_cast<player_laser_data_t *> (data));
        return 0;
    }    
    //////////////////////////
    // MAP OUTPUT INTERFACE //
    //////////////////////////
    // Is it a request for map meta-data?
  // Is it a request for map meta-data?
    else if(Message::MatchMessage(hdr, PLAYER_MSGTYPE_REQ, PLAYER_MAP_REQ_GET_INFO, this->map_addr))
    {
        PLAYER_MSG0(1,"Received PLAYER_MAP_REQ_GET_INFO request for the map interface");
        PLAYER_MSG6(1,"Scale:%.3f, Width:%u, Height:%u, Origin:(%.2f,%.2f,%.2f)", 
                    map_info.scale, map_info.width, map_info.height, 
                    map_info.origin.px, map_info.origin.py, map_info.origin.pa);

        this->Publish(this->map_addr, resp_queue, PLAYER_MSGTYPE_RESP_ACK, PLAYER_MAP_REQ_GET_INFO, (void*)&map_info, sizeof(map_info), NULL);
        return(0);
    }
    // Is it a request for a map tile?
    else if(Message::MatchMessage(hdr, PLAYER_MSGTYPE_REQ, PLAYER_MAP_REQ_GET_DATA, this->device_addr))
    {
        PLAYER_MSG0(1,"Received PLAYER_MAP_REQ_GET_DATA request for the map interface");
        
        player_map_data_t * map_req=(player_map_data_t*) data;

        // Can't declare a map tile on the stack (it's too big)
        size_t mapsize = sizeof(player_map_data_t);
        player_map_data_t* mapresp = (player_map_data_t*)calloc(1,mapsize);
        assert(mapresp);
        
        int i, j;
        int oi, oj, si, sj;

        // Construct reply
        oi = mapresp->col = map_req->col;
        oj = mapresp->row = map_req->row;
        si = mapresp->width = map_req->width;
        sj = mapresp->height = map_req->height;
        mapresp->data_count = mapresp->width * mapresp->height;
        mapresp->data = new int8_t [mapresp->data_count];
        // Grab the pixels from the map
        for(j = 0; j < sj; j++)
        {
            for(i = 0; i < si; i++)
            {
            //cur_map->cell(IntPoint(cur_map->getMapSizeX()-x-1, cur_map->getMapSizeY()-y-1));
                double v=this->cur_map->cell(IntPoint(j,i));
                if (v<0)
                    mapresp->data[i+j*si]=0; // unknown
                else if (v<0.5)
                    mapresp->data[i+j*si]=-1; // empty
                else
                    mapresp->data[i+j*si]=1; // occupied
/*                if(MAP_VALID(this, i + oi, j + oj))
                    mapresp->data[i + j * si] = this->mapdata[MAP_IDX(this, i+oi, j+oj)];
                else
                {
                    PLAYER_WARN2("requested cell (%d,%d) is offmap", i+oi, j+oj);
                    mapresp->data[i + j * si] = 0;
                }
*/          }
        }

        this->Publish(this->device_addr, resp_queue, PLAYER_MSGTYPE_RESP_ACK, PLAYER_MAP_REQ_GET_DATA, (void*)mapresp);
        delete [] mapresp->data;
        free(mapresp);
        return(0);
    }
    return 0;
}

////////////////////////////////////////////////////////////////////////////////
// Process new odometry data
void SlamDriver::ProcessOdom(player_msghdr_t* hdr, player_position2d_data_t &odom_data)
{

    // Cache the new odometric pose, velocity, and stall info
    this->cur_input_robot_pose.px=odom_data.pos.px;
    this->cur_input_robot_pose.py=odom_data.pos.py;
    this->cur_input_robot_pose.pa=odom_data.pos.pa;
    
    PLAYER_MSG3(4,"Received odometry [%f,%f,%f]", cur_input_robot_pose.px, cur_input_robot_pose.py, 180.0f/M_PI*cur_input_robot_pose.pa);

    // Assemble a new position message and publish it
    player_position2d_data_t new_odometry;
    new_odometry.pos.px=this->cur_input_robot_pose.px;
    new_odometry.pos.py=this->cur_input_robot_pose.py;
    new_odometry.pos.pa=this->cur_input_robot_pose.pa;
    // Copy the velocity values from the input position interface
    new_odometry.vel.px=odom_data.vel.px;
    new_odometry.vel.py=odom_data.vel.py;
    new_odometry.vel.pa=odom_data.vel.pa;
    new_odometry.stall=odom_data.stall;
  
    this->Publish(this->position_addr, PLAYER_MSGTYPE_DATA, PLAYER_POSITION2D_DATA_STATE, (void*)&new_odometry,sizeof(new_odometry),NULL); //&time);
//    player_msghdr_t newhdr = *hdr;
//    this->Publish(&newhdr, (void *)&new_odometry);
}


////////////////////////////////////////////////////////////////////////////////
// Process new laser data
void SlamDriver::ProcessLaser(player_msghdr_t* hdr, player_laser_data_t &laser_data)
{

    if (frontLaser == NULL)
    {   // Initialize laser sensor for gmapping
//        LaserData laserData = laserDataInPort.Get();

//        log(Info) << "Looping for laser data." << endlog();

//        while (laserData.getScanCount() == 0)
//        laserData = laserDataInPort.Get();

        this->laser_data.min_angle       = laser_data.min_angle;
        this->laser_data.max_angle       = laser_data.max_angle;
        this->laser_data.resolution      = laser_data.resolution;
        this->laser_data.max_range       = laser_data.max_range;
        this->laser_data.ranges_count    = laser_data.ranges_count;
        this->laser_data.intensity_count = laser_data.intensity_count;

        PLAYER_MSG6(1,"Got initial laser data: Min Angle:%.3f, Max Angle:%.3f, Resolution:%.3f, Max Range:%.3f, Scan Count:%u, Intensity Count:%u",
                      this->laser_data.min_angle, this->laser_data.max_angle,  this->laser_data.resolution, 
                      this->laser_data.max_range, this->laser_data.ranges_count, this->laser_data.intensity_count); 

        frontLaser = new RangeSensor("FLASER", this->laser_data.ranges_count, this->laser_data.resolution,  
                                     OrientedPoint(laser_pose3d.px,laser_pose3d.py,laser_pose3d.pyaw), 0, this->laser_data.max_range);

        frontLaser->updateBeamsLookup();

        SensorMap sensorMap;
        sensorMap.insert(make_pair("FLASER", frontLaser));

        // Set gmapping parameters
        gridSlamProcessor->setSensorMap(sensorMap);

        // Initialize gmapping
        gridSlamProcessor->init(particles, initial_map_x_min, initial_map_y_min, initial_map_x_max, initial_map_y_max, map_delta,
                                OrientedPoint(cur_input_robot_pose.px, cur_input_robot_pose.py, cur_input_robot_pose.pa)); // Put the current coordinates of the robot

        // This needs to be done, I dont know why, check with GMapping folks
        GridSlamProcessor* ap, *copy=gridSlamProcessor->clone();
        ap=gridSlamProcessor; gridSlamProcessor=copy; copy=ap;
    }
    else {
        // Convert laser and bot pose data to gmapping data structure
        RangeReading reading(frontLaser, hdr->timestamp); // time(NULL)
        reading.resize(frontLaser->beams().size());
        for (unsigned int i=0; i< laser_data.ranges_count; i++)
        {
           reading[i] = laser_data.ranges[i];
        }
        reading.setPose(OrientedPoint(this->cur_input_robot_pose.px, this->cur_input_robot_pose.py, this->cur_input_robot_pose.pa));

        // Process reading
        bool processed = gridSlamProcessor->processScan(reading);
        
        if (processed)
        {
            PLAYER_MSG0(1,"Updating SLAM map");
            // Get current best map index 
            unsigned int best_idx = gridSlamProcessor->getBestParticleIndex();

            // Convert it into a flat double map
            if (!cur_map)
                delete cur_map;
            cur_map = gridSlamProcessor->getParticles()[best_idx].map.toDoubleMap();

            // Get the corrected pose reported by slam
            OrientedPoint newPoseSlam = gridSlamProcessor->getParticles()[best_idx].pose;

            // TODO: check correcteness
//            BotPose newPose(newPoseSlam.x - initx, newPoseSlam.y - inity, newPoseSlam.theta);

            //////////////////////////
            // CAMERA INTERFACE     //
            //////////////////////////
            // Allocate the memory
            this->camera_data = reinterpret_cast<player_camera_data_t *>(malloc(sizeof(player_camera_data_t)));
            assert(this->camera_data);
            this->camera_data->width=cur_map->getMapSizeY();
            this->camera_data->height=cur_map->getMapSizeX();
            this->camera_data->bpp=24; // 256 values are sufficient. Use green color to denote unexplored regions
            this->camera_data->format=PLAYER_CAMERA_FORMAT_RGB888;//PLAYER_CAMERA_FORMAT_MONO8;//PLAYER_CAMERA_FORMAT_RGB888;
            this->camera_data->fdiv=0; // To get the probability divide by this value
            this->camera_data->compression=PLAYER_CAMERA_COMPRESS_RAW; // No compression
            this->camera_data->image_count=this->camera_data->width*this->camera_data->height*3; // Image size in pixels
            
            this->camera_data->image=new uint8_t[this->camera_data->image_count];
            assert(this->camera_data->image);
            memset(this->camera_data->image, 255, this->camera_data->image_count*sizeof(uint8_t));
            double v;
            for (int x=0; x<cur_map->getMapSizeX(); x++)
            {
                for (int y=0; y<cur_map->getMapSizeY(); y++)
                {
                    v=cur_map->cell(IntPoint(cur_map->getMapSizeX()-x-1, cur_map->getMapSizeY()-y-1));
                    // Unexplored space?
                    if (v<0)
                    {
                        this->camera_data->image[_3d_to_1d(x, y, 0, this->camera_data->height, this->camera_data->width, 3)]=0;
                        this->camera_data->image[_3d_to_1d(x, y, 1, this->camera_data->height, this->camera_data->width, 3)]=150;
                        this->camera_data->image[_3d_to_1d(x, y, 2, this->camera_data->height, this->camera_data->width, 3)]=0;
//                        PLAYER_MSG4(1,"Row:%u, Col: %u, Occupancy:%.2f, Color:%u", x, y, v, 750);
                         //continue;
//                    	PLAYER_MSG0(1,"JUSTIN IN THE HOUSE 1");
                    }
                    else
                    {
                    	uint8_t c = (uint8_t)(v*255.0f);
                        this->camera_data->image[_3d_to_1d(x, y, 0, this->camera_data->height, this->camera_data->width, 3)] = c;//(uint8_t)max(round(v*255.0f),1.);
                        this->camera_data->image[_3d_to_1d(x, y, 1, this->camera_data->height, this->camera_data->width, 3)] = c;//(uint8_t)max(round(v*255.0f),1.);
                        this->camera_data->image[_3d_to_1d(x, y, 2, this->camera_data->height, this->camera_data->width, 3)] = c;//(uint8_t)max(round(v*255.0f),1.);
                        //PLAYER_MSG4(1,"Row:%u, Col: %u, Occupancy:%.2f, Color:%u", x, y, v, uint8_t(v*255.0f));
                         // this->camera_data->image[_2d_to_1d(x, y, this->camera_data->height, this->camera_data->width)]=uint8_t(v*254);
                    
                    }
                }
            }
            if (write_map_to_file)
            {
                FILE * fp;
                fp = fopen( map_filename, "wb" );
                fprintf(fp,"P6\n%u %u\n255\n", this->camera_data->width, this->camera_data->height);
                fwrite((unsigned char*)this->camera_data->image, 1, this->camera_data->image_count, fp);
                fclose(fp);
            }

            PLAYER_MSG5(1,"Image Size: Width:%u, Height:%u, Bytes Per Pixel:%u, Divider:%u, Pixel Count:%u", 
                        this->camera_data->width, this->camera_data->height, this->camera_data->bpp, this->camera_data->fdiv, this->camera_data->image_count);            
            Publish(this->camera_addr, PLAYER_MSGTYPE_DATA, PLAYER_CAMERA_DATA_STATE, reinterpret_cast<void *>(this->camera_data), 0, &hdr->timestamp, false);
            // I assume that Publish() freed this data!
            this->camera_data = NULL;
            
            //////////////////////
            // MAP INTERFACE    //
            //////////////////////
            Point origin        = cur_map->map2world(IntPoint(0,0));
            map_info.scale      = this->map_delta;
            map_info.width      = cur_map->getMapSizeY();
            map_info.height     = cur_map->getMapSizeX();
            map_info.origin.px  = origin.x;
            map_info.origin.py  = origin.y;
            map_info.origin.pa  = 0.0f;

//                }
//            }            
/*            // Convert map to our coordinate system and IplImage
            IplImage* mapImg = cvCreateImage(cvSize(mymap->getMapSizeY(), mymap->getMapSizeX()),IPL_DEPTH_32F,1);

            cvSet(mapImg, cvScalar(2));
            // TODO: The (1,1) offset is from old code, not quite sure if its needed. Visually, seems better
            // but doesnt make sense to do that.

            for(int x = 1; x < mymap->getMapSizeX(); x++)
                for(int y = 1; y < mymap->getMapSizeY(); y++)
                {
                    double v=mymap->cell(IntPoint(mymap->getMapSizeX() - x - 1, mymap->getMapSizeY() - y - 1));

                    // Mark obstacles as 2 (used in PathFinder)
                    if (v < 0)
                        cvSet2D(mapImg, x, y, cvScalar(2));
                    else
                        cvSet2D(mapImg, x, y, cvScalar(v));
                }

            // This is ugly! We should either use GMapping structures or our own slam code.
            Point pmin=mymap->map2world(IntPoint(0,0));
            Point pmax=mymap->map2world(mymap->getMapSizeX()-1,mymap->getMapSizeY()-1);
            Point center=(pmax+pmin)*0.5;

            GridMap newMap(delta, mymap->getMapSizeY()/2 + center.y / delta - inity / delta,
            mymap->getMapSizeX()/2 + center.x / delta - initx / delta, 
            SharedIplImage(mapImg, IplImageDeleter));

            // Push updated slam
            slamMapViewerOutPort   .Push(newMap);
            slamMapExplorerOutPort .Push(newMap);
            slamMapAttentionOutPort.Set (newMap);

            // Set pose correction
            slamPoseCorrectionOutPort.Set(newPose-botPose);

            // Release the map
            delete mymap;
*/
            // This needs to be done, I dont know why, check with GMapping folks
            GridSlamProcessor* newProcessor = gridSlamProcessor->clone();
            delete gridSlamProcessor;
            gridSlamProcessor = newProcessor;
        }        
    }
}

////////////////////////////////////////////////////////////////////////////////
// Main function for device thread
void SlamDriver::Main() 
{
    // The main loop; interact with the device here
    for(;;)
    {
        
//        puts("Slam driver main... before waiting...");
        // Wait till we get new data
        this->Wait();
//        puts("Slam driver main... after waiting...");

        // test if we are supposed to cancel
        pthread_testcancel();

        if( this->InQueue->Empty() )
            return;

        // Process incoming messages.  ExampleDriver::ProcessMessage() is
        // called on each message.
        ProcessMessages();

        // TODO: forward the messages to the SLAM module
        
        // Publish data
/*        player_position2d_data_t data;

        memset(&data,0,sizeof(data));
        data.pos.px=1.0;
        data.pos.py=2.0;
        data.pos.px=0.01;
        // Push data out
        // TODO: Use the time when we received the last laser sensor for the time of the data
        this->Publish(this->position_addr, PLAYER_MSGTYPE_DATA, PLAYER_POSITION2D_DATA_STATE, (void*)&data,sizeof(data),NULL); //&time);
*/            
        // Sleep (you might, for example, block on a read() instead)
        usleep(50000);
    }
}

////////////////////////////////////////////////////////////////////////////////
// Extra stuff for building a shared object.

/* need the extern to avoid C++ name-mangling  */
extern "C" {
    int player_driver_init(DriverTable* table)
    {
        puts("Slam driver initializing");
        SlamDriver_Register(table);
        puts("Slam driver done");
        return(0);
    }
}

