/*
 *  SLAM player driver                     
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *

*/
#ifndef SLAMDRIVER_H
#define SLAMDRIVER_H

#include <unistd.h>
#include <string.h>

#include <iostream>
#include <fstream>

#include "gridslamprocessor.h"
using namespace GMapping;

#include <libplayercore/playercore.h>

////////////////////////////////////////////////////////////////////////////////
// The class for the SLAM driver
class SlamDriver : public Driver
{
    public:
        // Constructor; need that
        SlamDriver(ConfigFile* cf, int section);

        // Must implement the following methods.
        virtual int Setup();
        virtual int Shutdown();

        // This method will be invoked on each incoming message
        virtual int ProcessMessage(QueuePointer &resp_queue, player_msghdr * hdr, void * data);
        
    private:
        //////////////////
        // Functions    //
        //////////////////
        
        // Main function for device thread.
        virtual void Main();

        void ProcessOdom(player_msghdr_t* hdr, player_position2d_data_t &odom_data);
        void ProcessLaser(player_msghdr_t* hdr, player_laser_data_t &laser_data);

        //////////////////
        // Variables    //
        //////////////////
        // interfaces we might provide
        player_devaddr_t position_addr; // the new odometry position after registering the laser scans
        player_devaddr_t localize_addr;
        player_devaddr_t map_addr;
        player_devaddr_t camera_addr;  
        
        // Devices we require
        Device *odom;
        player_devaddr_t odom_addr;
        Device *laser;
        player_devaddr_t laser_addr;

        // GMapping objects
        RangeSensor       *frontLaser;
        GridSlamProcessor *gridSlamProcessor;

        // GMapping log file
        ofstream gmappingLogStream;
        
        ///////////////////////////////////////////////////////////////////////////////////////
        // Pose variables. We need to keep track of the input and output poses (in [m, m, rad])
        player_pose2d_t     cur_input_robot_pose; 
        player_pose2d_t     last_corrected_input_robot_pose;
        player_pose2d_t     cur_output_robot_pose;
        player_pose2d_t     last_corrected_output_robot_pose;

        //////////////////////
        // Laser variables  //
        //////////////////////
        player_pose3d_t         laser_pose3d; // With respect to the coordinate system centered on the midpoint of the front wheels
        player_bbox3d_t         laser_size3d; // The size of the laser sensor.
        player_laser_config_t   laser_config; // The configuration for our laser           
        player_laser_data_t     laser_data; // The data structure that contains some config parameters. Only updated at the reception of the first laser packet

        //////////////////////
        // Map variables    //
        //////////////////////
        Map<double, DoubleArray2D, false>*  cur_map;
        char*                               mapdata;
//        int                                 size_row, size_col;
        player_map_info_t                   map_info;
        
        //////////////////////
        // Camera variables //
        //////////////////////
        player_camera_data_t * camera_data;

        //////////////////////////////
        // Configuration variables
        int particles;
        // gfs measurement integration
        double angular_update, linear_update; // angular/linear threshold to update map
        // map resolution 
        double map_delta;
        // scan matcher 
//        double scan_matcher_max_range; // maximum valid range for urg laser=4m. We request that information from the laser device
//        double scan_matcher_max_urange; // (use up to)
        double scan_matcher_sigma; // scan matcher cell sigma for the greedy search
        int    scan_matcher_reg_score; // minimum score for registering a scan
        int    scan_matcher_iterations; // iterations
        double scan_matcher_crit_score; // critical score (leave this)
        double scan_matcher_max_move; // maximum move among two scans. This detects some corrupted logs
        bool   scan_matcher_autosize; // determine the map size by pre readoing the log
        
        double scan_matcher_linear_step; // linear search step (choose delta)
        double scan_matcher_angular_step; // angular search step, this is fine, depending on the odometry error and the update interval
        double scan_matcher_like_sigma; // sigma likelihood of 1 beam
        int    scan_matcher_like_skip; // beams to skip in the likelihood computation
        bool   scan_matcher_skip_matching; // do not perform scan matching before computing the statistics
        
        int    scan_matcher_kernel_size; // the higher the value the slower the filter, the better it can deal with noise, but the less precise and slower
        int    scan_matcher_gain_smooth; // gain for smoothing the likelihood
        double scan_matcher_resample_threshold; // when neff is below this value a resampling occurs
        // likelihood sampling
        double like_sampling_linear_range; // linear range
        double like_sampling_linear_step; // linear step
        double like_sampling_angular_range; // angular range
        double like_sampling_angular_step; // angular step
        // motion model parameters
        double motion_model_trans_fun_trans; // translation as a function of translation
        double motion_model_trans_fun_rot; // translation as a function of rotation
        double motion_model_rot_fun_trans; // rotation as a function of translation
        double motion_model_rot_fun_rot; // rotation as a function of rotation
        // odometry integration in proposal
        double linear_odometry_reliability; // linear odometry reliability
        double angular_odometry_reliability; // angular odometry reliability
        bool   consider_odometry_covariance; // consider odometry covariance
        // inital map params
        double initial_map_x_min; // minimum x for initial map
        double initial_map_y_min; // minimum y for initial map
        double initial_map_x_max; // maximum x for initial map
        double initial_map_y_max; // maximum y for initial map
        // initial robot position
        //     X
        //     ^
        //     |
        //Y <--+
//        double initial_robot_x; // initial robot position on the X axis
//        double initial_robot_y;  // initial robot position on the Y axis
//        double initial_robot_theta; // initial robot orientation
        
        bool   generate_map; // Generate a map or not?        
        bool   write_map_to_file; // Write the map image to a file?
        const char   *map_filename; // The image's filename        
};


#endif
